
/* This program is distributed as is -- no warranty of any kind is provided,
 * neither can the author, Franz Brauße, be held responsible for any condition
 * resulting from the execution of this software. */

#define _POSIX_C_SOURCE 2 /* getopt() */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>

#define KEY_ONE  0xa5
#define KEY_TWO  0x00
#define MAGIC   "xtRLE"

static int key1, key2;
static int verbose;
static uint32_t size;

static void decrypt(uint8_t *data, uint32_t size) {
	uint32_t i;
	uint8_t c, d;
	
	if (!size) return;
	
	c = data[0];
	for (i=1; ; i++) {
		d = data[i];
		if (d == c) {
			for (; i < size; i++) {
				d = data[i];
				data[i-1] = c ^ key2;
				if (d != c)
					break;
			}
		} else {
			data[i-1] = c ^ key1;
		}
		if (i >= size - 1) {
			data[size-1] = d ^ ((c == d) ? key2 : key1);
			break;
		}
		c = d;
	}
}

static uint8_t * decompress(FILE *in) {
	char magic[sizeof(MAGIC)];
	/* maybe 02 and 01 have something to do with single and multiple occurences? */
	uint32_t unknown;
	uint8_t *data = NULL, *data_ptr;
	unsigned num;
	int c;
	
	if (!fread(magic, sizeof(MAGIC), 1, in)) goto err_generic;
	if (!fread(&unknown, sizeof(unknown), 1, in)) goto err_generic;
	if (!fread(&size, sizeof(size), 1, in)) goto err_generic;
	
	if (memcmp(magic, MAGIC, sizeof(MAGIC))) {
		fprintf(stderr,
				"the 6 magic bytes of the supplied file don't match %s\n", MAGIC);
		goto err;
	}
	
	if (verbose)
		fprintf(stderr, "header read, decompressing %u bytes to memory\n", size);
	
	data = malloc(size);
	if (!data) {
		perror("allocating memory for scrambled, decompressed content");
		goto err;
	}
	
	data_ptr = data;
	while ((c = fgetc(in)) != EOF) {
		if (!c) {
			fseek(in, 0x0e + num, SEEK_SET);
		} else {
			num = c & 0x7f;
			if (c & 0x80) {
				/* num bytes plain data follow */
				if (!fread(data_ptr, num, 1, in))
					goto err_generic;
			} else {
				if ((c = fgetc(in)) == EOF) {
					fprintf(stderr,
							"short read: expected argument for run of %u bytes\n", num);
					goto err;
				}
				memset(data_ptr, c, num);
			}
			data_ptr += num;
		}
	}
	
	return data;
	
err_generic:
	fprintf(stderr, "error reading input file: %s\n", strerror(ferror(in)));
err:
	free(data);
	return NULL;
}

int main(int argc, char **argv) {
	const char *nin = NULL, *nout = NULL;
	key1 = KEY_ONE;
	key2 = KEY_TWO;
	verbose = 0;
	FILE *fin, *fout;
	uint8_t *data = NULL;
	int ret = EXIT_SUCCESS;
	char *endptr;
	
	int opt;
	while ((opt = getopt(argc, argv, ":vh1:2:")) != -1) {
		switch (opt) {
			case '1':
				key1 = strtol(optarg, &endptr, 0);
				if (*endptr) {
					fprintf(stderr, "not a valid number: '%s', error interpreting '%s'\n", optarg, endptr);
					return EXIT_FAILURE;
				}
				break;
			case '2':
				key2 = strtol(optarg, &endptr, 0);
				if (*endptr) {
					fprintf(stderr, "not a valid number: '%s', error interpreting '%s'\n", optarg, endptr);
					return EXIT_FAILURE;
				}
				break;
			case 'v': verbose++; break;
			case ':':
				fprintf(stderr, "option '-%c' needs an argument\n", optopt);
				return EXIT_FAILURE;
			case '?': /* fall through */
			case 'h':
				fprintf(stderr, "usage: %s [-1 key1] [-2 key2] [encrypted.dat] [output]\n", argv[0]);
				return EXIT_FAILURE;
		}
	}
	
	if (optind < argc) nin = argv[optind++];
	if (optind < argc) nout = argv[optind++];
	
	fin = (nin) ? fopen(nin, "r") : stdin;
	if (!fin) {
		perror("opening input file");
		return EXIT_FAILURE;
	}
	fout = (nout) ? fopen(nout, "w") : stdout;
	if (!fout) {
		perror("opening output file");
		fclose(fin);
		return EXIT_FAILURE;
	}
	
	if (verbose)
		fprintf(stderr, "decompressing input to memory\n");
	
	data = decompress(fin);
	fclose(fin);
	if (!data) {
		ret = EXIT_FAILURE;
		goto end;
	}
	
	if (verbose)
		fprintf(stderr,
				"using key 0x%02x for single and 0x%02x for multiple occurences\n",
				key1, key2);
	
	decrypt(data, size);
	if (!fwrite(data, size, 1, fout)) {
		fprintf(stderr, "error writing output: %s\n", strerror(ferror(fout)));
		ret = EXIT_FAILURE;
	}
	
end:
	free(data);
	fclose(fout);
	return ret;
}
